#extension GL_OES_EGL_image_external : enable

// This shader is intended to run on a single-pixel framebuffer in order to
// compute an average color over the region defined by uFilterBounds. It is an
// extension of the simpler region_filter_fragment.glsl shader, with the ability
// to handle lighting estimation from the camera feed when the surface albedo of
// the filtered region is unknown. More info:
// go/living-surfaces-relighting-albedo.

precision mediump float;

uniform samplerExternalOES sTexture;
uniform vec4 uFilterBounds;  // Contains [xMin, xMax, yMin, yMax].
uniform mat4 uPreviewM;
uniform mat4 uTextureM;

// The constants below were derived manually from multiple test runs under
// varying lighting conditions and using different target images with varying
// albedo distributions.

// Saturation maximum for full-frame color average. Applied per sample.
const float kMaxFrameSaturation = 0.25;

// Saturation maximum for filtered region color average. Applied per sample.
const float kMaxRegionSaturation = 0.125;

// Degree to which a filtered region can be darker than overall frame
// average. Higher values handle nonuniform lighting better but tend to
// overdarken results when the target image's albedo is dark.
const float kRegionDarknessLocality = 0.5;

// This is the albedo color that is assumed in order to infer the returned
// lighting color. Due to heuristics in this shader that minimize the influence
// of dark, saturated colors in the camera feed, a single value for
// kPresumedAlbedo works reasonably well across a wide range of lighting
// scenarios and actual albedos.
const vec3 kPresumedAlbedo = vec3(0.55);

// The uFilterBounds region is sampled with kRegionReso x kRegionReso samples.
const float kRegionReso = 8.0;

// The full frame is sampled with kFrameReso x kFrameReso samples.
const float kFrameReso = 8.0;

const float kEpsilon = 0.001;

// Basic color space conversions; see
// http://lolengine.net/blog/2013/07/27/rgb-to-hsv-in-glsl.
vec3 rgbToHsv(vec3 color) {
  const vec4 kRgbToHsvK = vec4(0.0, -1.0 / 3.0, 2.0 / 3.0, -1.0);
  const float kENegTen = 1.0e-10;
  vec4 p = mix(vec4(color.bg, kRgbToHsvK.wz), vec4(color.gb, kRgbToHsvK.xy),
               step(color.b, color.g));
  vec4 q = mix(vec4(p.xyw, color.r), vec4(color.r, p.yzx), step(p.x, color.r));
  float d = q.x - min(q.w, q.y);
  return vec3(abs(q.z + (q.w - q.y) / (6.0 * d + kENegTen)),
    d / (q.x + kENegTen), q.x);
}

vec3 hsvToRgb(vec3 color) {
  const vec4 kHsvToRgbK = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
  vec3 p = abs(fract(color.xxx + kHsvToRgbK.xyz) * 6.0 - kHsvToRgbK.www);
  return color.z * mix(kHsvToRgbK.xxx, clamp(p - kHsvToRgbK.xxx, 0.0, 1.0),
                       color.y);
}

// Determines sample weight for a color in HSV space. The weight favors bright,
// unsaturated colors, as these tend to incorporate a mostly white albedo and
// are most likely to be representative of the lighting environment.
float hsvToSampleWeight(vec3 hsv) {
  return clamp(hsv.z * (1.0 - hsv.y), kEpsilon, 1.0);
}

// Saturation clamping in HSV space.
vec3 hsvToSatClampedHsv(vec3 hsv, float max_sat) {
  return vec3(hsv.x, min(hsv.y, max_sat), hsv.z);
}

// Saturation clampingin HSV space, output as RGB.
vec3 hsvToSatClampedRgb(vec3 hsv, float max_sat) {
  return hsvToRgb(hsvToSatClampedHsv(hsv, max_sat));
}

/** Maps normalized canvas coordinates into the external texture space. */
vec2 mapToExternalTexture(float fx, float fy) {
  vec2 view = vec2(fx, fy) * 2.0 - 1.0;
  vec2 deview = vec2(view.x / uPreviewM[0][0], view.y / uPreviewM[1][1]);
  vec2 tex = (uTextureM * vec4((deview + 1.0) * 0.5, 0, 1)).xy;
  return tex;
}

// Returns representative color of the full frame. This is an average color of
// multiple saturation-clamped samples covering the whole camera feed.
vec3 getFrameColor() {
  vec3 sum = vec3(0.0);
  vec2 frameMin = clamp(mapToExternalTexture(0.0, 0.0), 0.0, 1.0);
  vec2 frameMax = clamp(mapToExternalTexture(1.0, 1.0), 0.0, 1.0);
  for (float y = 0.5; y < kFrameReso; y += 1.0) {
    float fy = mix(frameMin.y, frameMax.y, y / kFrameReso);
    for (float x = 0.5; x < kFrameReso; x += 1.0) {
      float fx = mix(frameMin.x, frameMax.x, x / kFrameReso);
      vec3 rgb = texture2D(sTexture, vec2(fx, fy)).rgb;
      vec3 hsv = rgbToHsv(rgb);
      sum += hsvToSatClampedRgb(hsv, kMaxFrameSaturation);
    }
  }
  return sum / (kFrameReso * kFrameReso);
}

void main() {
  vec3 color = vec3(0.0);
  vec3 frameRgb = getFrameColor();
  vec3 frameHsv = rgbToHsv(frameRgb);
  float frameWeight = hsvToSampleWeight(frameHsv);
  float frameDiv = 1.0 / max(kEpsilon, frameHsv.z);

  vec2 regionMin = clamp(
      mapToExternalTexture(uFilterBounds[0], uFilterBounds[2]), 0.0, 1.0);
  vec2 regionMax = clamp(
      mapToExternalTexture(uFilterBounds[1], uFilterBounds[3]), 0.0, 1.0);
  for (float y = 0.5; y < kRegionReso; y += 1.0) {
    float fy = mix(regionMin.y, regionMax.y, y / kRegionReso);
    for (float x = 0.5; x < kRegionReso; x += 1.0) {
      float fx = mix(regionMin.x, regionMax.x, x / kRegionReso);
      vec3 regionRgb = texture2D(sTexture, vec2(fx, fy)).rgb;
      vec3 regionHsv = rgbToHsv(regionRgb);
      float regionWeight = hsvToSampleWeight(regionHsv);
      regionHsv = hsvToSatClampedHsv(regionHsv, kMaxRegionSaturation);
      regionRgb = hsvToRgb(regionHsv);

      float lerpToRegion = regionWeight / (regionWeight + frameWeight);
      float frameScale =
          mix(1.0, regionHsv.z * frameDiv, kRegionDarknessLocality);
      vec3 scaledFrameRgb = frameScale * frameRgb;
      regionRgb = mix(scaledFrameRgb, regionRgb, lerpToRegion);
      color += regionRgb;
    }
  }
  color *= 1.0 / (kRegionReso * kRegionReso * kPresumedAlbedo);
  gl_FragColor = vec4(color, 1.0);
}
